function [u,itrs] = REST_TV_L1_U_QMM(b,blur_k,mu,eps_R,eps_F,u,itrs_th,itrs_rel_chg_th,itrs_debugs,n_sigma,u_true)
%
% Compute approximate solutions (that is, minimizers) of the
% unconstrained TV-L1 variational model: 
%
% u*(mu) = argmin J(u;mu),   J(u;mu) = TV(u) + mu ||K u - b||_1  
%             u
%
% for the non-blind restoration, that is denoising + deblurring (non-blind 
% means that the blur operator is assumed to be known), of grayscale images 
% using the QMM (iterative) optimization algorithm, as illustrated in [1-3].
%
% [1] P. Rodriguez and B. Wohlberg, Efficient Minimization Method for a 
%     Generalized Total Variation Functional, IEEE Transactions on Image 
%     Processing, 2009, vol.18(2), pp.322-332, 2009.


% INPUTS:
    % inputs defining the unconstrained TV-L1 cost function to be minimized:
        % b         --> observed corrupted (blurred and noisy) image
        % blur_k    --> blur PSF (kernel of spatial 2D convolution)
        % mu        --> regularization parameter (positive real scalar)        
    % inputs defining the QMM (iterative) algorithm:
        % eps_R     --> smoothing parameter for the regularizer (positive real scalar)
        % eps_F     --> smoothing parameter for the fidelity (positive real scalar)
        % u         --> initial iterate (usually, but not necessarily, the observed corrupted image b)
        % QMM stopping criteria:
            % itrs_th         --> threshold: maximum number of QMM iterations
            % itrs_rel_chg_th --> threshold: iterates relative change
    % (optional) inputs defining eventual "debugs" 
        % itrs_debugs --> flag: 0->no debug; 1->debug
        % n_sigma     --> noise standard deviation
        % u_true      --> original uncorrupted image
        
% We remark the following:
%   - all images are stored as matrices, that is they are not vectorized!
%   - periodic boundary conditions are assumed for the unknown clean image 
%     to restore, that is both the blur and the first-order derivatives 
%     near the image boundaries are computed according to this assumption!
%     2D convolutions are performed by using the Matlab command imfilter.
    
% --------------------
% initialize algorithm
% --------------------

    % set stopping criteria for inner pcg iterations
    pcg_iters_th = 500; % maximum number of iterations     for pcg at each outer iteration
    rel_res_sf   = 5;   % relative residual scaling factor for pcg at each outer iteration
    
    % extract image dimensions
    size_b = size(b); h = size_b(1); w = size_b(2); d = h * w;
    
    % initialize optimization
    eps_R2 = eps_R^2; eps_F2 = eps_F^2; lambda = 1 / mu;
    K      = @(x) imfilter(x,blur_k,'circular','conv');
    KT     = @(x) K(x);
    Dhu    = Dh(u); Dvu = Dv(u); Ku_b = K(u) - b;
    u_vec  = u(:);
            
% --------------------------------------
% (eventually) compute/store/show debugs
% --------------------------------------
if ( itrs_debugs == 1 )
    % allocate arrays where storing iterations-based debug quantities
    J_regs        = zeros(1,itrs_th+1);
    J_fids        = zeros(1,itrs_th+1);
    rel_chgs      = zeros(1,itrs_th);
    Ku_b_means    = zeros(1,itrs_th+1);
    taus          = zeros(1,itrs_th+1);
    snrs          = zeros(1,itrs_th+1);
    ssims         = zeros(1,itrs_th+1);
    % compute/initialize and store iterations-based debug quantities
    J_regs(1)     = sum( sqrt(Dhu(:).^2 + Dvu(:).^2) );
    J_fids(1)     = mu * sum( abs(Ku_b(:)) );
    Ku_b_means(1) = mean( Ku_b(:) );
    taus(1)       = sqrt( sum( Ku_b(:).^2 ) ) / (sqrt(d)*n_sigma);
    snrs(1)       = compute_snr(u,u_true);
    ssims(1)      = ssim(u,u_true);  
    % Matlab command window debugs
    fprintf('\n---------------------------------------------------------');
    fprintf('\nTV-L1-U-QMM (mu = %7.2f) ITERATIONS-BASED DEBUGS:',mu);
    fprintf('\n---------------------------------------------------------');
    fprintf('\nIT %04d:  RC = %9.7f, J = %13.6f, tau* = %6.4f, [ISNR,ISSIM] = [%6.3f , %6.4f]',...
        0,0,(J_regs(1)+J_fids(1)),taus(1),snrs(1)-snrs(1),ssims(1)-ssims(1));
end
    
% -------------------------
% carry out QMM iterations
% -------------------------

% initialize iteration index and stopping criteria flags
itrs        = 0;   
stop_flags  = [0,0];

% while stopping criteria are not satisfied...iterate
while ( sum(stop_flags) == 0 )
    
    % update iteration index
    itrs = itrs + 1;
    
    % MAJORIZE: compute the majorization weights
    WR = 1./sqrt( eps_R2 +  Dhu.^2 + Dvu.^2 );
    WF = 1./sqrt( eps_F2 + Ku_b.^2 );
        
    % MINIMIZE: solve the symmetric positive definite linear system
        % compute the current right-hand side vector
        rhs_vec   = vec( KT( WF .* b ) );
        % compute the relative residual of the current iterate u
        rel_res   = norm( A(u_vec,size_b,K,KT,WR,WF,lambda) - rhs_vec ) / norm(rhs_vec);
        % depending on the relative residual, set the pcg tolerance
        pcg_tol   = rel_res / rel_res_sf;
        % store u_vec_old for future computation of iterates relative change
        u_vec_old = u_vec;  
        % compute the new iterate by solving the linear system by pcg
        [u_vec, flg, rlr, pit] = pcg(@(x) A(x,size_b,K,KT,WR,WF,lambda), ...
                                 rhs_vec,pcg_tol,pcg_iters_th,[],[],u_vec);
        % reshape u_vec into matrix u
        u = reshape(u_vec,size_b);
        
    % compute u relative change (for stopping criterion)
    rel_chg = norm(u_vec - u_vec_old) / norm(u_vec_old);
    % check stopping criteria
    if ( itrs == itrs_th ),           stop_flags(1) = 1; end
    if ( rel_chg < itrs_rel_chg_th ), stop_flags(2) = 1; end
    % compute quantities used in the next iteration
    Dhu = Dh(u); Dvu = Dv(u); Ku_b = K(u) - b;
    % (eventually) compute/store/show debugs
    if ( itrs_debugs == 1 )
        % compute and store iterations-based debug quantities
        J_regs(itrs+1)     = sum( sqrt(Dhu(:).^2 + Dvu(:).^2) );
        J_fids(itrs+1)     = mu * sum( abs(Ku_b(:)) );
        rel_chgs(itrs)     = rel_chg;
        Ku_b_means(itrs+1) = mean( Ku_b(:) );
        taus(itrs+1)       = sqrt( sum( Ku_b(:).^2 ) ) / (sqrt(d)*n_sigma);
        snrs(itrs+1)       = compute_snr(u,u_true);
        ssims(itrs+1)      = ssim(u,u_true);
        % Matlab command window debugs
        fprintf('\nIT %04d:  RC = %9.7f, J = %13.6f, tau* = %6.4f, [ISNR,ISSIM] = [%6.3f , %6.4f]',...
            itrs,rel_chgs(itrs),(J_regs(itrs+1)+J_fids(itrs+1)),taus(itrs+1),snrs(itrs+1)-snrs(1),ssims(itrs+1)-ssims(1));
    end
    
end

% one of the two stopping criteria satisfied --> end iterations  
% and, before returning the restored image u and the total number 
% of performed QMM iterations itrs to the main script, show 
% (eventually if itrs_debugs = 1) some iterations-based debug quantities

% show iterations-based debug quantities
if ( itrs_debugs == 1 )
    
    % first, shrink arrays where we stored iterations-based debug quantities
    J_regs        = J_regs(1:itrs+1);
    J_fids        = J_fids(1:itrs+1);
    rel_chgs      = rel_chgs(1:itrs);
    Ku_b_means    = Ku_b_means(1:itrs+1);
    taus          = taus(1:itrs+1);
    snrs          = snrs(1:itrs+1);
    ssims         = ssims(1:itrs+1);
    
    % then show all debugs
    fprintf('\n');
    
    plots_x_min = 0;
    plots_x_max = itrs;
    
    figure('Name','Figure 70:  TV-L1-U-QMM:   ITERATIONS-BASED DEBUGS','NumberTitle','off');
    set(gcf,'Position',get(0,'ScreenSize'));
    
    subplot(2,3,1)
    plot(0:itrs,J_regs,'b');
    xlabel('k  (QMM iteration index)');
    ylabel('$\mathrm{TV}\left(u^{(k)}\right) = \sum_i \left\| (\nabla u)_i \right\|_2$','interpreter','latex');
    plot1_y_min = min( J_regs(2:end) );
    plot1_y_max = max( J_regs(2:end) );
    plot1_y_min = plot1_y_min - 0.05 * (plot1_y_max - plot1_y_min);
    plot1_y_max = plot1_y_max + 0.05 * (plot1_y_max - plot1_y_min);
    axis([plots_x_min,plots_x_max,plot1_y_min,plot1_y_max]);
    
    subplot(2,3,2)
    plot(0:itrs,J_fids,'k');
    xlabel('k  (QMM iteration index)');
    ylabel('$\mu \,\, \mathrm{L}_1\left(u^{(k)};b,K\right) = \mu \,\, \left\|Ku^{(k)}-b\right\|_1$','interpreter','latex');
    plot2_y_min = min( J_fids(2:end) );
    plot2_y_max = max( J_fids(2:end) );
    plot2_y_min = plot2_y_min - 0.05 * (plot2_y_max - plot2_y_min);
    plot2_y_max = plot2_y_max + 0.05 * (plot2_y_max - plot2_y_min);
    axis([plots_x_min,plots_x_max,plot2_y_min,plot2_y_max]);
    
    subplot(2,3,3)
    plot(0:itrs,J_regs + J_fids,'r');
    xlabel('k  (QMM iteration index)');
    ylabel('$J\left(u^{(k)};\mu\right) = \mathrm{TV}\left(u^{(k)}\right) + \mu \, \mathrm{L}_1\left(u^{(k)};b,K\right)$','interpreter','latex');
    plot3_y_min = min( J_regs(2:end) + J_fids(2:end) );
    plot3_y_max = max( J_regs(2:end) + J_fids(2:end) );
    plot3_y_min = plot3_y_min - 0.05 * (plot3_y_max - plot3_y_min);
    plot3_y_max = plot3_y_max + 0.05 * (plot3_y_max - plot3_y_min);
    axis([plots_x_min,plots_x_max,plot3_y_min,plot3_y_max]);
    
    subplot(2,3,4)
    plot(0:(itrs-1),rel_chgs,'b');
    xlabel('k  (QMM iteration index)');
    ylabel('$\delta^{(k)} = \left\| u^{(k+1)}-u^{(k)} \right\|_2 / \left\|u^{(k)}\right\|_2$','interpreter','latex');
    plot4_y_min = min( rel_chgs(2:end) );
    plot4_y_max = max( rel_chgs(2:end) );
    plot4_y_min = plot4_y_min - 0.05 * (plot4_y_max - plot4_y_min);
    plot4_y_max = plot4_y_max + 0.05 * (plot4_y_max - plot4_y_min);
    axis([plots_x_min,plots_x_max,plot4_y_min,plot4_y_max]);
    
    subplot(2,3,5)
    plot(0:itrs,taus,'b');
    xlabel('k  (QMM iteration index)');
    ylabel('$\tau^{(k)} = \left\|K u^{(k)} - b\right\|_2 / (\sqrt{d} \: \sigma_n)$','interpreter','latex');
    plot5_y_min = min( taus(2:end) );
    plot5_y_max = max( taus(2:end) );
    plot5_y_min = plot5_y_min - 0.05 * (plot5_y_max - plot5_y_min);
    plot5_y_max = plot5_y_max + 0.05 * (plot5_y_max - plot5_y_min);
    axis([plots_x_min,plots_x_max,plot5_y_min,plot5_y_max]);
    
    subplot(2,3,6)    
    yyaxis left;
    plot(0:itrs,snrs-snrs(1));
    xlabel('k  (QMM iteration index)'); 
    ylabel('$ISNR\left(u_{true},u^{(k)}\right)$','interpreter','latex');
    plot6_y_min = min( snrs(2:end)-snrs(1) );
    plot6_y_max = max( snrs(2:end)-snrs(1) );
    plot6_y_max = plot6_y_max + 0.05 * (plot6_y_max - plot6_y_min);
    axis([plots_x_min,plots_x_max,plot6_y_min,plot6_y_max]);
    yyaxis right;
    plot(0:itrs,ssims-ssims(1));
    xlabel('k  (QMM iteration index)');
    ylabel('$ISSIM\left(u_{true},u^{(k)}\right)$','interpreter','latex');
    plot6_y_min = min( ssims(2:end)-ssims(1) );
    plot6_y_max = max( ssims(2:end)-ssims(1) );
    plot6_y_max = plot6_y_max + 0.05 * (plot6_y_max - plot6_y_min);
    axis([plots_x_min,plots_x_max,plot6_y_min,plot6_y_max]);
    
end

%-------------------------
% NESTED FUNCTIONS
%-------------------------
function v = vec(u)
  v = u(:);
end
function Dhu = Dh(u)
    Dhu  = [ u(:,2:end) - u(:,1:(end-1)) , u(:,1) - u(:,end) ];
end
function Dvu = Dv(u)
    Dvu  = [ u(2:end,:) - u(1:(end-1),:) ; u(1,:) - u(end,:) ]; 
end
function DhTu = DhT(u)
    DhTu = [ u(:,end) - u(:,1) , -diff(u,1,2) ];
end
function DvTu = DvT(u)
    DvTu = [ u(end,:) - u(1,:) ; -diff(u,1,1) ];
end
% Compute u = A v, with v a vectorized image (since called by pcg function) 
% and A the s.p.d. matrix A = KT*WF*K + lambda * (DhT*WR*Dh + DvT*WR*Dv)
function u_vec = A(v_vec,v_size,K,KT,WR,WF,lambda)
    % reshape column vector v to the original matrix form V
    v = reshape(v_vec,v_size);
    % compute the output in matrix form u
    u = KT(WF.*K(v)) + lambda * ( DhT(WR.*Dh(v)) + DvT(WR.*Dv(v)) );
    % finally, reshape matrix u into output column vector u_vec
    u_vec = u(:);
end


end


